parser grammar fsqParser;
 
options {
  language = Java;
  tokenVocab = fsqLexer;
}

@header {
  package de.uni_tuebingen.sfb.lichtenstein.formulas.parsing;

  import de.uni_tuebingen.sfb.lichtenstein.formulas.*;
}

/**
 * One formula on each line whitespace allowed in between.  Newlines should be, but aren't, ignored by the lexer.
 */
formulaList : WHITESPACE? ( formula WHITESPACE? )* EOF ;

/**
 * A formula is enclosed by braces and returns a Formula.
 */
formula returns [Formula result] 
	: OPEN body CLOSE
		{ $result = $body.result; }
	;

/**
 * There are 6 kinds of formulas.
 */
body returns [Formula result]
	: labelFormula
		{ $result = $labelFormula.result; }
	| atomic
		{ $result = $atomic.result; }
	| unary
		{ $result = $unary.result; }
	| binary
		{ $result = $binary.result; }
	| n_ary
		{ $result = $n_ary.result; }
	| quantor
		{ $result = $quantor.result; }
	;

// TODO: create separate atomic formulas for each label formula
/**
 * A label formula can be either of five.
 */
labelFormula returns [Formula result]
	: FUNCTION WHITESPACE firstOrderVariable WHITESPACE label
		{ $result = new Inclusion($firstOrderVariable.result, $label.result); }
	| CATEGORY WHITESPACE firstOrderVariable WHITESPACE label
		{ $result = new Inclusion($firstOrderVariable.result, $label.result); }
	| WORD WHITESPACE firstOrderVariable WHITESPACE label
		{ $result = new Inclusion($firstOrderVariable.result, $label.result); }
	| LEMMA WHITESPACE firstOrderVariable WHITESPACE label
		{ $result = new Inclusion($firstOrderVariable.result, $label.result); }
	| MORPHOLOGY WHITESPACE firstOrderVariable WHITESPACE label
		{ $result = new Inclusion($firstOrderVariable.result, $label.result); }
	;

/**
 * The different atomic relations.
 */
atomic returns [Formula result]
	: equality
		{ $result = $equality.result; }
	| properDominance
		{ $result = $properDominance.result; }
	| immediateDominance
		{ $result = $immediateDominance.result; }
	| dominance
		{ $result = $dominance.result; }
	| immediatePrecedence
		{ $result = $immediatePrecedence.result; }
	| precedence
		{ $result = $precedence.result; }
	| containment
		{ $result = $containment.result; }
	;

/**
 * The equality of two nodes or sets??
 */
equality returns [Formula result]
	: EQUALITY WHITESPACE 
		  ( firstfirst=firstOrderVariable WHITESPACE secondfirst=firstOrderVariable
			{ $result = new FirstOrderEquality($firstfirst.result, $secondfirst.result); } )
		| ( firstsecond=secondOrderVariable WHITESPACE secondsecond=secondOrderVariable
			{ $result = new SecondOrderEquality($firstsecond.result, $secondsecond.result); } )
	;

/**
 * 
 */
properDominance returns [Formula result]
	: PROPER_DOMINANCE WHITESPACE first=firstOrderVariable WHITESPACE second=firstOrderVariable
		{ $result = new ProperDominance($first.result, $second.result); }
	;

/**
 * 
 */
immediateDominance returns [Formula result]
	: IMMEDIATE_DOMINANCE WHITESPACE first=firstOrderVariable WHITESPACE second=firstOrderVariable
		{ $result = new ImmediateDominance($first.result, $second.result); }
	;

/**
 * 
 */
dominance returns [Formula result]
	: DOMINANCE WHITESPACE first=firstOrderVariable WHITESPACE second=firstOrderVariable
		{ $result = new Dominance($first.result, $second.result); }
	;

/**
 * 
 */
immediatePrecedence returns [Formula result]
	: IMMEDIATE_PRECEDENCE WHITESPACE first=firstOrderVariable WHITESPACE second=firstOrderVariable
		{ $result = new ImmediatePrecedence($first.result, $second.result); }
	;

/**
 * 
 */
precedence returns [Formula result]
	: PRECEDENCE WHITESPACE first=firstOrderVariable WHITESPACE second=firstOrderVariable
		{ $result = new Precedence($first.result, $second.result); }
	;

/**
 * 
 */
containment returns [Formula result]
	: CONTAINMENT WHITESPACE first=firstOrderVariable WHITESPACE second=secondOrderVariable
		{ $result = new Inclusion($first.result, $second.result); }
	;

/**
 * Negation.
 */
unary returns [Formula result]
	: NEGATION WHITESPACE arg=formula
		{ $result = new Negation($arg.result); }
	;

/**
 * Implication or biimplication.
 */
binary returns [Formula result]
	: IMPLICATION WHITESPACE first=formula WHITESPACE second=formula
		{ $result = new Implication($first.result, $second.result); }
	| EQUIVALENCE WHITESPACE first=formula WHITESPACE second=formula
		{ $result = new Equivalence($first.result, $second.result); }
	;

/**
 * Conjunction or disjunction.
 */
n_ary returns [Formula result]
	: CONJUNCTION args=n_aryArguments
		{ $result = new Conjunction(args.toArray(new Formula[args.size()])); }
	| DISJUNCTION args=n_aryArguments
		{ $result = new Disjunction(args.toArray(new Formula[args.size()])); }
	;

n_aryArguments returns [List<Formula> result]
@init{ $result = new ArrayList<Formula>(); }
	: ( WHITESPACE arg=formula { $result.add(arg); } )+
	;

/**
 * There are first order and second order quantors.
 */
quantor returns [Formula result]
	: FIRST_ORDER_EXISTENTIAL WHITESPACE firstOrderVariable WHITESPACE formula
		{ $result = new FirstOrderExistentialQuantification($firstOrderVariable.result, $formula.result); }
	| FIRST_ORDER_UNIVERSAL WHITESPACE firstOrderVariable WHITESPACE formula
		{ $result = new FirstOrderUniversalQuantification($firstOrderVariable.result, $formula.result); }
	| SECOND_ORDER_EXISTENTIAL WHITESPACE secondOrderVariable WHITESPACE formula
		{ $result = new SecondOrderExistentialQuantification($secondOrderVariable.result, $formula.result); }
	| SECOND_ORDER_UNIVERSAL WHITESPACE secondOrderVariable WHITESPACE formula
		{ $result = new SecondOrderUniversalQuantification($secondOrderVariable.result, $formula.result); }
	;

// variables
firstOrderVariable returns [FirstOrderVariable result]
	: IDENTIFIER
		{ $result = FirstOrderVariable.getInstance($IDENTIFIER.text); }
	;

secondOrderVariable returns [SecondOrderVariable result]
	: IDENTIFIER
		{ $result = SecondOrderVariable.getInstance($IDENTIFIER.text); }
	;

label returns [Predicate result]
	: IDENTIFIER
		{ $result = Predicate.getInstance($IDENTIFIER.text); }
	;
